import numpy as np
import gym

class Env:
  def __init__(self, name, render=False):
    self.name = name
    self.env = gym.make(name)
    self.S = self.env.observation_space
    self.A = self.env.action_space

  def step(self, a):
    ś, r, t, info = self.env.step(a)
    if t and 'TimeLimit.truncated' in info:
      truncated = info['TimeLimit.truncated']
    else:
      truncated = False
    return ś.astype('f4', copy=False), r.astype('f4', copy=False), t, truncated
  def reset(self):
    return self.env.reset().astype('f4', copy=False)
  def seed(self, i):
    self.env.seed(i)
  def close(self):
    self.env.close()

class AbsorbingWrapper(gym.ObservationWrapper):
  def __init__(self, env):
    super().__init__(env)
    obs_space = self.observation_space
    self.observation_space = gym.spaces.Box(
        shape=(obs_space.shape[0] + 1,),
        low=obs_space.low[0],
        high=obs_space.high[0])
    self.absorbing_state = np.zeros(self.observation_space.shape, dtype=np.float32)
    self.absorbing_state[-1] = 1
  def observation(self, obs):
    return np.concatenate([obs, [0]], -1)

class AbsEnv:
  def __init__(self, name, render=False):
    self.name = name
    self.env = AbsorbingWrapper(gym.make(name))
    self.absorbing_state = self.env.absorbing_state
    self.S = self.env.observation_space
    self.A = self.env.action_space

  def step(self, a):
    ś, r, t, info = self.env.step(a)
    if t and 'TimeLimit.truncated' in info:
      truncated = info['TimeLimit.truncated']
    else:
      truncated = False
    return ś.astype('f4', copy=False), r.astype('f4', copy=False), t, truncated
  def reset(self):
    return self.env.reset().astype('f4', copy=False)
  def seed(self, i):
    self.env.seed(i)
  def close(self):
    self.env.close()
